'''
<2020.06.23>
Classify available models to regression and classification.

reference: https://scikit-learn.org/stable/index.html
'''

REGRESSION = ['Regressor', 'Ridge', 'Lasso', 'ElasticNet', 'Lars', 'SVR', 'OrthogonalMatchingPursuit']
CLASSIFICATION = ['Classifier', 'Logistic', 'SVC', 'NB', 'Perceptron']
ONEVS = ['OneVsRest', 'OneVsOne']


